@@ -83,6 +83,10 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
             return new PopupMenu.PopupMenuItem(opts.label);
         };
 
+        // TODO: When no windows exist on the active workspace, but do on another,
+        // we should detect those cases and offer workspace options if they are pinned. Otherwise,
+        // user needs to switch workspaces just to switch the window back. This should behave this way
+        // when showAllWorkspaces is disabled as well since its a UX problem.
         if (hasWindows) {
             // Monitors
             if (Main.layoutManager.monitors.length > 1) {
@@ -108,6 +112,9 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                     item = createMenuItem({label: _('Only on this workspace')});
                     this.signals.connect(item, 'activate', () => {
                         this.groupState.lastFocused.unstick();
+                        // Always index windows from all workspaces while showAllWorkspaces is enabled
+                        if (this.state.settings.showAllWorkspaces) return;
+                        this.state.removingWindowFromWorkspaces = true;
                         this.state.trigger('removeWindowFromOtherWorkspaces', this.groupState.lastFocused);
                     });
                     this.addMenuItem(item);
@@ -128,17 +135,18 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                         });
                     };
                     for (let i = 0; i < length; i++) {
-                        // Make the index a local letiable to pass to function
+                        // Make the index a local variable to pass to function
                         let j = i;
                         let name = Main.workspace_names[i] ? Main.workspace_names[i] : Main._makeDefaultWorkspaceName(i);
-                        let ws = createMenuItem({label: _(name)});
+                        let menuItem = createMenuItem({label: _(name)});
+                        let ws = this.groupState.lastFocused.get_workspace();
 
-                        if (i === this.state.currentWs) {
-                            ws.setSensitive(false);
+                        if (ws && i === ws.index()) {
+                            menuItem.setSensitive(false);
                         }
 
-                        connectWorkspaceEvent(ws, j);
-                        item.menu.addMenuItem(ws);
+                        connectWorkspaceEvent(menuItem, j);
+                        item.menu.addMenuItem(menuItem);
                     }
                 }
             }
@@ -274,19 +282,17 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
         // Pin/unpin, shortcut handling
         if (!isWindowBacked) {
-            if (this.state.settings.showPinned !== FavType.none) {
-                let label, icon;
-                if (this.groupState.isFavoriteApp) {
-                    label = _('Unpin from Panel');
-                    icon = 'unpin';
-                } else {
-                    label = _('Pin to Panel');
-                    icon = 'pin';
-                }
-                this.pinToggleItem = createMenuItem({label, icon});
-                this.signals.connect(this.pinToggleItem, 'activate', (...args) => this.toggleFavorite(...args));
-                this.addMenuItem(this.pinToggleItem);
+            let label, icon;
+            if (this.groupState.isFavoriteApp) {
+                label = _('Unpin from Panel');
+                icon = 'unpin';
+            } else {
+                label = _('Pin to Panel');
+                icon = 'pin';
             }
+            this.pinToggleItem = createMenuItem({label, icon});
+            this.signals.connect(this.pinToggleItem, 'activate', (...args) => this.toggleFavorite(...args));
+            this.addMenuItem(this.pinToggleItem);
             if (this.state.settings.autoStart) {
                 let label = this.groupState.autoStartIndex !== -1 ? _('Remove from Autostart') : _('Add to Autostart');
                 item = createMenuItem({label: label, icon: 'insert-object'});
@@ -305,7 +311,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         if (hasWindows) {
             let metaWindowActor = this.groupState.lastFocused.get_compositor_private();
             // Miscellaneous
-            if (metaWindowActor.opacity !== 255) {
+            if (metaWindowActor && metaWindowActor.opacity !== 255) {
                 item = createMenuItem({label: _('Restore to full opacity')});
                 this.signals.connect(item, 'activate', () => metaWindowActor.set_opacity(255));
                 this.addMenuItem(item);
@@ -351,6 +357,9 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                 });
                 this.addMenuItem(item);
                 // Close all
+                // TODO: We should detect if windows from this group are on another workspace
+                // and close windows across all workspaces while showAllWorkspaces is enabled.
+                // Ditto for 'Close others'.
                 item = createMenuItem({label: _('Close all'), icon: 'application-exit'});
                 this.signals.connect(item, 'activate', () => {
                     if (!this.groupState.isFavoriteApp) {
@@ -488,6 +497,21 @@ class HoverMenuController extends PopupMenu.PopupMenuManager {
         this.groupState.disconnect(this.connectId);
         super.destroy();
     }
+
+    _onMenuOpenState(menu, open) {
+        if (open) {
+            if (this._activeMenu && this._activeMenu.isChildMenu(menu)) {
+                this._menuStack.push(this._activeMenu);
+            }
+            this._activeMenu = menu;
+        } else {
+            if (this._menuStack.length > 0) {
+                this._activeMenu = this._menuStack.pop();
+                if (menu.sourceActor)
+                    this._didPop = true;
+            }
+        }
+    }
 }
 
 class WindowThumbnail {
@@ -614,6 +638,7 @@ class WindowThumbnail {
 
     onLeave() {
         this.entered = false;
+        this.stopClick = false;
         this.actor.remove_style_pseudo_class('selected');
         this.onFocusWindowChange();
         this.button.set_opacity(0);
@@ -659,6 +684,8 @@ class WindowThumbnail {
         this.metaWindow.delete(global.get_current_time());
         if (!this.groupState.metaWindows || this.groupState.metaWindows.length <= 1) {
             this.groupState.trigger('hoverMenuClose');
+        } else {
+            this.groupState.trigger('checkShouldClose');
         }
     }
 
@@ -698,10 +725,11 @@ class WindowThumbnail {
         } else {
             this.metaWindowActor = this.metaWindow.get_compositor_private();
         }
-        if (this.metaWindowActor) {
+        if (this.metaWindowActor && !this.metaWindowActor.is_finalized()) {
             this.signals.connect(this.metaWindowActor, 'size-changed', () => this.refreshThumbnail());
 
             let windowTexture = this.metaWindowActor.get_texture();
+            if (!windowTexture) return;
             let [width, height] = windowTexture.get_size();
             let scale = Math.min(1.0, thumbnailWidth / width, thumbnailHeight / height) * global.ui_scale;
             width = Math.round(width * scale);
@@ -798,7 +826,10 @@ class WindowThumbnail {
     }
 
     hoverPeek(opacity) {
-        if (!this.state.settings.enablePeek || this.state.overlayPreview || this.state.scrollActive) {
+        if (!this.state.settings.enablePeek
+            || this.state.overlayPreview
+            || this.state.scrollActive
+            || (this.metaWindowActor && this.metaWindowActor.is_finalized())) {
             return;
         }
         if (!this.metaWindowActor) {
@@ -857,6 +888,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         super._init.call(this, groupState.trigger('getActor'), state.orientation, 0.5);
         this.state = state;
         this.groupState = groupState;
+        this.setCustomStyleClass("grouped-window-list-thumbnail-menu");
 
         this.connectId = this.groupState.connect({
             hoverMenuClose: () => {
@@ -864,11 +896,26 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
                 this.groupState.set({thumbnailMenuEntered: false});
                 this.close();
             },
+            checkShouldClose: () => {
+                // This is called after a close button is clicked. When the menu size changes, it can leave the cursor
+                // outside the menu bounds, and no leave event will be able to correct this situation where the menu is
+                // dangling open and only closable upon hovering over it again. This checks if the cursor is hovering
+                // over the menu and closes it if not.
+                setTimeout(() => {
+                    let [x, y, mask] = global.get_pointer();
+                    let draggedOverActor = global.stage.get_actor_at_pos(Clutter.PickMode.ALL, x, y);
+                    let parent = draggedOverActor.get_parent();
+                    if (!(parent instanceof St.Widget)) {
+                        this.close(true);
+                    }
+                }, 500);
+            },
             addThumbnailToMenu: (win) => {
                 if (this.isOpen) {
-                    setTimeout(() => this.addThumbnail(win), 0);
+                    this.addThumbnail(win);
                     return;
                 }
+
                 this.queuedWindows.push(win);
             },
             removeThumbnailFromMenu: (win) => {
@@ -887,6 +934,30 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
                 this.setVerticalSetting();
                 if (isOpen) this.open(true);
             },
+            fileDrag: ({fileDrag}) => {
+                if (fileDrag) {
+                    // When a drag operation from another app is started, no events fire, so we have to grab the
+                    // cursor, find the actor by coordinates, and then look up the thumbnail actor. Do this on a
+                    // 50ms loop until the menu closes so we continue getting data in the absence of events.
+                    this.interval = setInterval(() => {
+                        let [x, y, mask] = global.get_pointer();
+                        let draggedOverActor = global.stage.get_actor_at_pos(Clutter.PickMode.ALL, x, y);
+                        if (draggedOverActor instanceof Meta.ShapedTexture) {
+                            this.groupState.set({fileDrag: false});
+                            this.close(true);
+                            return;
+                        }
+                        each(this.appThumbnails, function(thumbnail) {
+                            if (thumbnail.thumbnailActor === draggedOverActor) {
+                                Main.activateWindow(thumbnail.metaWindow, global.get_current_time());
+                                return false;
+                            }
+                        });
+                    }, 50);
+                } else if (this.interval) {
+                    clearInterval(this.interval);
+                }
+            }
         });
 
         this.appThumbnails = [];
@@ -894,6 +965,12 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         this.fullyRefreshThumbnails();
     }
 
+    addQueuedThumbnails() {
+        if (this.queuedWindows.length === 0) return;
+        each(this.queuedWindows, (win) => this.addThumbnail(win));
+        this.queuedWindows = [];
+    }
+
     onButtonPress() {
         if (this.state.settings.onClickThumbs && this.box.get_children().length > 1) {
             return;
@@ -918,10 +995,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             timeout = this.state.settings.thumbTimeout;
         }
 
-        if (this.queuedWindows.length > 0) {
-            each(this.queuedWindows, (win) => this.addThumbnail(win));
-            this.queuedWindows = [];
-        }
+        this.addQueuedThumbnails();
 
         if (actor != null) {
             this.groupState.set({thumbnailMenuEntered: this.isOpen});
@@ -964,6 +1038,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             this.groupState.tooltip.set_text(this.groupState.appName);
             this.groupState.tooltip.show();
         } else {
+            if (force || this.state.settings.onClickThumbs) this.addQueuedThumbnails();
             this.state.set({thumbnailMenuOpen: true});
             super.open(this.state.settings.animateThumbs);
         }
@@ -988,6 +1063,10 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         for (let i = 0; i < this.appThumbnails.length; i++) {
             this.appThumbnails[i].destroyOverlayPreview();
         }
+
+        if (this.groupState.fileDrag) {
+            this.groupState.set({fileDrag: false});
+        }
     }
 
     onKeyPress(actor, e) {
@@ -1064,7 +1143,6 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             this.destroyThumbnails();
         }
         this.addWindowThumbnails(this.groupState.metaWindows);
-        this.setStyleOptions();
     }
 
     destroyThumbnails() {
@@ -1126,29 +1204,6 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         }
     }
 
-    setStyleOptions() {
-        if (this.willUnmount || !this.box) return;
-
-        // The styling cannot be set correctly unless the menu is closed. Fortunately this
-        // can be closed and reopened too quickly for the user to notice.
-        let {isOpen} = this;
-        if (isOpen) this.close(true);
-
-        this.box.show();
-        this.box.style = null;
-
-        let thumbnailTheme = this.box.peek_theme_node();
-        let padding = thumbnailTheme ? thumbnailTheme.get_horizontal_padding() : null;
-        let thumbnailPadding = padding && (padding > 1 && padding < 21) ? padding : 10;
-        this.box.style = `padding: ${thumbnailPadding / 2}px`;
-        let boxTheme = this.box.peek_theme_node();
-        padding = boxTheme ? boxTheme.get_vertical_padding() : null;
-        let boxPadding = padding && padding > 0 ? padding : 3;
-        this.box.style = `padding: ${boxPadding}px;`;
-
-        if (isOpen) this.open(true);
-    }
-
     setVerticalSetting() {
         if (this.state.orientation === St.Side.TOP || this.state.orientation === St.Side.BOTTOM) {
             this.box.vertical = this.groupState.verticalThumbs || this.state.settings.verticalThumbs;
@@ -1169,13 +1224,8 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         for (let i = 0; i < this.appThumbnails.length; i++) {
             if (this.appThumbnails[i]) {
                 this.appThumbnails[i].refreshThumbnail();
-                // Make sure Clutter updates, otherwise setStyleOptions is called afterwards,
-                // and will be calculating styles from old actor values because it closes the menu
-                // to avoid incorrect padding values.
-                this.appThumbnails[i].thumbnailActor.realize();
             }
         }
-        this.setStyleOptions();
     }
 
     destroy() {
